package no.statnett.ecp.brs;

import no.statnett.ecp.utils.ArtemisConsoleAPI;
import no.statnett.ecp.utils.Const;
import no.statnett.ecp.utils.EcpBroker;
import no.statnett.ecp.utils.Options;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.StandardOpenOption;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.time.LocalDateTime;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

public class BrokerShell {

    public static final String VERSION = "1.0.1";

    // OPTIONS
    private static int consumerCount = -1;
    private static String queueNames = null;
    private static String searchString = null;
    private static int sortAscending = 0;
    private static int sortDescending = 0;
    private static int limitCount = -1;
    private static boolean interactive = false; // default
    private static boolean purge = false; // default
    private static boolean remove = false;
    private static Map<String, Login> extraLogins = null;
    private static File jsonFile = null;
    private static List<String> filterHistory = new ArrayList<>();


    public static void main(String[] args) throws IOException, NoSuchAlgorithmException, KeyManagementException {
        List<String> mandatoryArgs = parseOptions(args);
        if (mandatoryArgs.size() != 1) {
            System.out.println("Expected exactly 1 (one) mandatory argument, either URL/PATH to broker, but received " + mandatoryArgs.size());
            usage();
            System.exit(1);
        }
        Map<String, String> config = EcpBroker.findBrokerConfig(mandatoryArgs.get(0), false);
        extraLogins.remove(config.get("host")); // Remove the main host from the extraHosts list - to avoid double listing
        List<QueueInfo> qiList = listQueues(config, extraLogins);
        print(qiList, jsonFile);
        if (remove) {
            delete(qiList, config);
        } else if (purge) {
            purge(qiList, config);
        } else if (interactive) {
            interactive(qiList, config);
        }
    }


    private static List<String> parseOptions(String[] initialArgs) throws IOException {
        List<String> initArgsList = new ArrayList<>(Arrays.asList(initialArgs));
        consumerCount = Options.parseInt(initArgsList, "-c", 0, Integer.MAX_VALUE, -1);
        queueNames = Options.parseString(initArgsList, "-q");
        searchString = Options.parseString(initArgsList, "-s");
        sortAscending = Options.parseInt(initArgsList, "-a", 1, 6, 0);
        sortDescending = Options.parseInt(initArgsList, "-d", 1, 6, 0);
        limitCount = Options.parseInt(initArgsList, "-l", 1, Integer.MAX_VALUE, 0);
        if (sortAscending > 0) // -a takes precedence
            sortDescending = 0;
        interactive = Options.parseBoolean(initArgsList, "-i");
        purge = Options.parseBoolean(initArgsList, "-p");
        remove = Options.parseBoolean(initArgsList, "-r");
        if (interactive) { // -i takes precedence
            purge = false;
            remove = false;
        } else if (purge) { // -p takes precedence over -r
            remove = false;
        }
        String loginFileStr = Options.parseString(initArgsList, "-h");
        if (loginFileStr != null) {
            File loginFile = Options.checkFile(loginFileStr, false);
            extraLogins = parseLoginFile(loginFile);
        }
        String jsonFilename = Options.parseString(initArgsList, "-o");
        if (jsonFilename != null) {
            jsonFile = Options.checkFile(jsonFilename, true);
        }
        return initArgsList;
    }

    private static Map<String, Login> parseLoginFile(File loginFile) throws IOException {
        // The loginFile has the following format, the 3 last fields are optional:
        // hostname [user password protocol]
        Map<String, Login> loginMap = new HashMap<>();
        for (String line : Files.readAllLines(loginFile.toPath())) {
            if (line.startsWith("#"))
                continue;
            String[] parts = line.split(" ");
            if (parts.length == 4) {
                loginMap.put(parts[0], new Login(parts[0], parts[1], parts[2], parts[3]));
            } else if (parts.length == 1) {
                loginMap.put(parts[0], new Login(parts[0], null, null, null));
            }
        }
        return loginMap;
    }





    /* MAIN METHODS FOR LISTING, PURGING, DELETING QUEUES */

    private static void print(List<QueueInfo> sortedQIList) throws IOException {
        print(sortedQIList, null);
    }

    private static void print(List<QueueInfo> sortedQIList, File jsonFile) throws IOException {
        System.out.println(String.format("%-4s | %-25s | %-50s | %9s | %8s | %11s | %11s ", "#", "Host (1)", "Queue (2)", "Size (3)", "Cons (4)", "Enqueue (5)", "Dequeue (6)") + "(" + LocalDateTime.now().format(Const.localTmsSec) + ")");
        System.out.println("-".repeat(136));

        int i = 1;
        for (QueueInfo qi : sortedQIList) {
            StringBuilder sb = new StringBuilder();
            sb.append(String.format("%-4d |", i));
            sb.append(String.format(" %-25s |", qi.getTrimmedHost()));
            sb.append(String.format(" %-50s |", qi.getName()));
            sb.append(String.format(" %9d |", qi.getSize()));
            sb.append(String.format(" %8d |", qi.getConsumers()));
            sb.append(String.format(" %11d |", qi.getEnqueueCount()));
            sb.append(String.format(" %11d", qi.getDequeueCount()));
            System.out.println(sb);
            i++;
        }

        if (jsonFile != null) {
            for (QueueInfo qi : sortedQIList) {
                StringBuilder sb = new StringBuilder();
                sb.append("{ \"statustime\": \"").append(LocalDateTime.now().format(Const.localTmsSec)).append("\", ");
                sb.append("\"host\": ").append(format(40, qi.getHost()));
                sb.append("\"queue\": ").append(format(50, qi.getName()));
                sb.append("\"size\": ").append(format(5, qi.getSize()));
                sb.append("\"consumers\": ").append(format(2, qi.getConsumers()));
                sb.append("\"enqueue\": ").append(format(6, qi.getEnqueueCount()));
                sb.append("\"dequeue\": ").append(formatNoComma(6, qi.getDequeueCount()));
                sb.append("}\n");
                Files.writeString(jsonFile.toPath(), sb.toString(), StandardOpenOption.APPEND);
            }
        }
    }

    private static List<QueueInfo> listQueues(Map<String, String> config, Map<String, Login> extraLogins) throws NoSuchAlgorithmException, IOException, KeyManagementException {
        List<QueueInfo> QIList = findQueueInfo(prot(config, null), config.get("host"), config.get("port"), config.get("user"), config.get("password")); // Find the main host first
        for (Login login : extraLogins.values()) {
            String user = login.getUser() == null ? config.get("user") : login.getUser();
            String password = login.getPassword() == null ? config.get("password") : login.getPassword();
            String protocol = login.getProtocol() == null ? prot(config, null) : login.getProtocol();
            QIList.addAll(findQueueInfo(protocol, login.getHost(), config.get("port"), user, password));
        }
        List<QueueInfo> trimmedHostnameList = trimHostNames(QIList);
        List<QueueInfo> filteredByQueueName = filterByQueueNames(trimmedHostnameList, queueNames);
        List<QueueInfo> filteredBySearchString = filterBySearchString(filteredByQueueName, searchString);
        List<QueueInfo> filteredByConsumerCount = filterByConsumerCount(filteredBySearchString, consumerCount);
        List<QueueInfo> sortedQIList = sortQueueInfo(filteredByConsumerCount, sortAscending, sortDescending);
        List<QueueInfo> limitedQIList = limitCount > 0 ? sortedQIList.subList(0, Math.min(limitCount, sortedQIList.size())) : sortedQIList;
        return limitedQIList;
    }

    private static void purge(List<QueueInfo> limitedQIList, Map<String, String> config) {
        int messagesPurged = 0;
        int executionsOk = 0;
        int failed = 0;
        System.out.print("Starting to purge.");
        Map<String, String> broker2Encoded = new HashMap<>();
        for (QueueInfo qi : limitedQIList) {
            try {
                if (broker2Encoded.get(qi.getHost()) == null) { // The same broker on the same host, no need to look up again
                    String brokerNameEncoded = ArtemisConsoleAPI.retrieveBrokerNameURLEncoded(prot(config, qi), qi.getHost(), config.get("port"), user(config, qi), pass(config, qi));
                    broker2Encoded.put(qi.getHost(), brokerNameEncoded);
                }
                ArtemisConsoleAPI.ACAPIResult result = ArtemisConsoleAPI.purgeQueue(prot(config, qi), qi.getHost(), config.get("port"), broker2Encoded.get(qi.getHost()), user(config, qi), pass(config, qi), qi.getName());
                if (result.getStatus() != null && result.getStatus().equals("200")) {
                    executionsOk++;
                    messagesPurged += Integer.parseInt(result.getValue());
                } else {
                    System.out.println("Failed to purge queue: " + qi.getName() + " on host: " + qi.getHost() + " - " + result.getJson());
                    failed++;
                }
                System.out.print(".");
            } catch (Exception e) {
                System.out.println("Error occurred: During purging/removeAllMessages of queue: " + qi.getName() + " on host " + qi.getHost() + ": " + e.getMessage());
                failed++;
            }
        }
        String failedStr = failed > 0 ? "Failed to purge " + failed + " queues." : "";
        System.out.println("\nPurged completed: " + messagesPurged + " messages purged from " + executionsOk + " out of " + limitedQIList.size() + " queues listed. " + failedStr);
    }

    private static void delete(List<QueueInfo> limitedQIList, Map<String, String> config) throws NoSuchAlgorithmException, IOException, KeyManagementException {
        int executionsOk = 0;
        int failed = 0;
        System.out.print("Starting to delete queues.");
        Map<String, String> broker2Encoded = new HashMap<>();
        for (QueueInfo qi : limitedQIList) {
            try {
                if (broker2Encoded.get(qi.getHost()) == null) { // The same broker on the same host, no need to look up again
                    String brokerNameEncoded = ArtemisConsoleAPI.retrieveBrokerNameURLEncoded(prot(config, qi), qi.getHost(), config.get("port"), user(config, qi), pass(config, qi));
                    broker2Encoded.put(qi.getHost(), brokerNameEncoded);
                }
                boolean success = ArtemisConsoleAPI.deleteQueue(prot(config, qi), qi.getHost(), config.get("port"), broker2Encoded.get(qi.getHost()), user(config, qi), pass(config, qi), qi.getName());
                if (success) {
                    executionsOk++;
                } else {
                    failed++;
                }
                System.out.print(".");
            } catch (Exception e) {
                System.out.println("Error occurred: During deletion of queue: " + qi.getName() + " on host " + qi.getHost() + ": " + e.getMessage());
                failed++;
            }
        }
        String failedStr = failed > 0 ? "Failed to delete " + failed + " queues." : "";
        System.out.println("\nDelete is completed: Delete " + executionsOk + " out of " + limitedQIList.size() + " queues listed. " + failedStr);
    }

    private static String user(Map<String, String> config, QueueInfo qi) {
        if (qi == null)
            return config.get("user");
        return extraLogins.get(qi.getHost()) != null && extraLogins.get(qi.getHost()).getUser() != null ? extraLogins.get(qi.getHost()).getUser() : config.get("user");
    }

    private static String pass(Map<String, String> config, QueueInfo qi) {
        if (qi == null)
            return config.get("password");
        return extraLogins.get(qi.getHost()) != null && extraLogins.get(qi.getHost()).getPassword() != null ? extraLogins.get(qi.getHost()).getPassword() : config.get("password");
    }

    private static String prot(Map<String, String> config, QueueInfo qi) {
        // The protocol may or may not have been specified by the user input
        String protocol = config.get("suggest") == null ? "https" : config.get("suggest").equals("tls") ? "https" : "http";
        // Override the above protocol if specified in the extraLogins-file
        if (qi != null) {
            if (extraLogins.get(qi.getHost()) != null && extraLogins.get(qi.getHost()).getProtocol() != null)
                protocol = extraLogins.get(qi.getHost()).getProtocol();
        }
        return protocol;
    }

    private static void interactive(List<QueueInfo> qiList, Map<String, String> config) throws NoSuchAlgorithmException, IOException, KeyManagementException {
        System.out.println("BrokerShell for Artemis v" + VERSION + " - Interactive mode. Enter h for help");
        Scanner scanner = new Scanner(System.in);
        String input = "";
        do {
            System.out.print(">");
            input = scanner.nextLine();
            if (!input.trim().isEmpty()) {
                if (input.equalsIgnoreCase("h"))
                    help();
                else if (input.equalsIgnoreCase("z"))
                    showSettings();
                else if (input.equalsIgnoreCase("v")) {
                    qiList = listQueues(config, extraLogins); // updates the qiList and prints it
                    print(qiList);
                } else if (input.equalsIgnoreCase("f")) {
                    for (int i = 0; i < filterHistory.size(); i++) {
                        System.out.println((i + 1) + ": " + filterHistory.get(i));
                    }
                } else if (input.startsWith("c")) {
                    Integer tmp = getInt(input);
                    if (tmp == null)
                        continue;
                    if (tmp < 0)
                        System.out.println("Filtering by consumer number is turned off, will match ALL consumer count");
                    else
                        System.out.println("Consumer count filter is set to " + tmp);
                    consumerCount = tmp;
                } else if (input.startsWith("s")) {
                    searchString = null;
                    if (input.length() > 1) {
                        searchString = input.substring(1).trim();
                        if (searchString.trim().isEmpty())
                            searchString = null;
                    }
                    if (searchString == null)
                        System.out.println("Filtering by search string is turned off, will match ALL queue listing");
                    else {
                        filterHistory.add(input);
                        System.out.println("Search string filter is set to '" + searchString + "' (will be negated: " + searchString.startsWith("!") + ")");
                    }
                } else if (input.startsWith("q")) {
                    queueNames = null;
                    if (input.length() > 1) {
                        queueNames = input.substring(1).trim();
                        if (queueNames.trim().isEmpty())
                            queueNames = null;
                    }
                    if (queueNames == null)
                        System.out.println("Filtering by queue names is turned off, will match ALL queue listing");
                    else {
                        filterHistory.add(input);
                        System.out.println("Queue names filter is set to " + queueNames);
                    }
                } else if (input.startsWith("a")) {
                    Integer tmp = getInt(input);
                    if (tmp == null)
                        continue;
                    if (tmp > 6) {
                        System.out.println("Sorting by column " + tmp + " is not supported, only 1-6, command is ignored");
                    } else if (tmp < 1) {
                        if (sortDescending > 0) {
                            System.out.println("Ascending sort is turned off, will go back to descending sort on column " + sortDescending);
                        } else {
                            System.out.println("Ascending sort is turned off, will show default listing");
                        }
                        sortAscending = 0;
                    } else {
                        sortAscending = tmp;
                        sortDescending = 0;
                        System.out.println("Sorting by column " + tmp + " in ascending order");
                    }
                } else if (input.startsWith("d")) {
                    Integer tmp = getInt(input);
                    if (tmp == null)
                        continue;
                    if (tmp > 6) {
                        System.out.println("Sorting by column " + tmp + " is not supported, only 1-6, command is ignored");
                    } else if (tmp < 1) {
                        if (sortAscending > 0) {
                            System.out.println("Descending sort is turned off, but has no effect since ascending sort of column " + sortAscending + " takes precedence");
                        } else {
                            System.out.println("Descending sort is turned off, will show default listing");
                        }
                        sortDescending = 0;
                    } else {
                        sortDescending = tmp;
                        sortAscending = 0;
                        System.out.println("Sorting by column " + tmp + " in descending order");
                    }
                } else if (input.startsWith("l")) {
                    Integer tmp = getInt(input);
                    if (tmp == null)
                        continue;
                    if (tmp < 1) {
                        limitCount = 0;
                        System.out.println("Limit is turned off, will show all queues that matches the filter-settings");
                    } else {
                        limitCount = tmp;
                        System.out.println("Limit is set to " + tmp + " queues");
                    }
                } else if (input.startsWith("p")) {
                    Integer tmp = getInt(input);
                    if (tmp == null)
                        continue;
                    if (tmp == -1) {
                        System.out.println("Purging all queues in the printed list, since number is -1");
                        purge(qiList, config);
                    } else if (tmp > 0 && tmp <= qiList.size()) {
                        System.out.println("Purging queue number " + tmp);
                        List<QueueInfo> selectedQIList = new ArrayList<>();
                        selectedQIList.add(qiList.get(tmp - 1));
                        purge(selectedQIList, config);
                    } else {
                        System.out.println("Invalid queue number: " + tmp + ", command ignored");
                    }
                } else if (input.startsWith("r")) {
                    Integer tmp = getInt(input);
                    if (tmp == null)
                        continue;
                    if (tmp == -1) {
                        System.out.println("Removing all queues in the printed list, since number is -1");
                        delete(qiList, config);
                    } else if (tmp > 0 && tmp <= qiList.size()) {
                        System.out.println("Removing queue number " + tmp);
                        List<QueueInfo> selectedQIList = new ArrayList<>();
                        selectedQIList.add(qiList.get(tmp - 1));
                        delete(selectedQIList, config);
                    } else {
                        System.out.println("Invalid queue number: " + tmp + ", command ignored");
                    }

                }
            }


        } while (!input.equalsIgnoreCase("x"));
        scanner.close();
        System.out.println("Exiting BrokerShell");
    }

    private static Integer getInt(String input) {
        try {
            if (input.length() > 1)
                return Integer.parseInt(input.substring(1));
            else {
                System.out.println("Expected a number after the command, command ignored");
                return null;
            }
        } catch (NumberFormatException e) {
            System.out.println("Invalid number: " + input.substring(1) + ", command ignored");
            return null;
        }
    }

    private static void help() {
        System.out.println("Interactive mode commands:");
        System.out.println("   h          - help");
        System.out.println("   x          - exit");
        System.out.println(" LISTINGS:");
        System.out.println("   v          - view queues");
        System.out.println("   f          - show filter (search/queues) history");
        System.out.println("   z          - show filter/sort settings");
        System.out.println(" FILTERS:");
        System.out.println("   c<number>  - lists queues with this number of consumers. Remove filter by setting to -1");
        System.out.println("   s <string> - lists lines matching this regexp. Prefixing with ! will negate search. Remove filter by 's' alone");
        System.out.println("   q <queues> - lists queues with these names (comma-separated). Remove filter by setting q without argument.");
        System.out.println("   a<number>  - Ascending sort by column 1-host 2-name 3-size 4-consumers 5-enqueue 6-dequeue. Remove sort by setting to 0.");
        System.out.println("   d<number>  - Descending sort by column 1-host 2-name 3-size 4-consumers 5-enqueue 6-dequeue. Remove sort by setting to 0.");
        System.out.println("   l<number>  - limit output to this number of lines (after filter/sorting is applied). Remove limit by setting to 0");
        System.out.println(" ACTIONS - TAKE CARE!:");
        System.out.println("   p<number>  - purge queue (number is queue number). Purge ALL queues in the printed list if number is -1");
        System.out.println("   r<number>  - remove queues (number is queue number). Remove ALL queues in the printed list if number is -1");
    }

    private static void showSettings() {
        String consumerFilter = consumerCount < 0 ? "<ALL>" : String.valueOf(consumerCount);
        String queueNamesFilter = queueNames == null ? "<ALL>" : queueNames;
        String searchStringFilter = searchString == null ? "<ALL>" : searchString.startsWith("!") ? searchString + " (negated:true)" : searchString + " (negated:false)";
        String sortSetting = ", Sort: {Nothing specified}";
        if (sortAscending > 0) {
            sortSetting = ", Sort: {Ascending by column " + sortAscending + "}";
        } else if (sortDescending > 0) {
            sortSetting = ", Sort: {Descending by column " + sortDescending + "}";
        }
        String limitSetting = ", View-limit: ";
        if (limitCount < 1) {
            limitSetting += "{No limitation - will show all that matches the filters}";
        } else {
            limitSetting += "{Will show the *first* " + limitCount + " queues}";
        }
        System.out.println("Filters: {no_of_consumers=" + consumerFilter + ", queues=" + queueNamesFilter + ", reg-exp/search-string=" + searchStringFilter + "}" + sortSetting + limitSetting);
    }






    /* SUB-METHODS SUPPORTING THE ABOVE MAIN METHODS */

    public static List<QueueInfo> findQueueInfo(String prot, String host, String port, String username, String password) throws NoSuchAlgorithmException, IOException, KeyManagementException {
        String brokerNameEncoded = ArtemisConsoleAPI.retrieveBrokerNameURLEncoded(prot, host, port, username, password);
        String json = ArtemisConsoleAPI.retrieveQueueInfoJSON(prot, host, port, brokerNameEncoded, username, password);
        int dataArrPosStart = json.indexOf("data\\\":[") + 7;
        int dataArrPosEnd = json.indexOf("]", dataArrPosStart);

        String dataArrJson = json.substring(dataArrPosStart, dataArrPosEnd);
        dataArrJson = dataArrJson.replaceAll("\\\\", "");

        List<QueueInfo> queueInfoList = new ArrayList<>();
        for (String elementJson : dataArrJson.split("},\\{")) {
            // Each element looks like this (except first is prefixed with "[{" and last is postfixed with "}".
            // "id":"7","name":"ExpiryQueue","address":"ExpiryQueue","filter":"","durable":"true","paused":"false","temporary":"false","purgeOnNoConsumers":"false","consumerCount":"0","maxConsumers":"-1","autoCreated":"false","user":"","routingType":"ANYCAST","messagesAdded":"0","messageCount":"0","messagesAcked":"0","messagesExpired":"0","deliveringCount":"0","messagesKilled":"0","directDeliver":"false","exclusive":"false","lastValue":"false","lastValueKey":"","scheduledCount":"0","groupRebalance":"false","groupRebalancePauseDispatch":"false","groupBuckets":"-1","groupFirstKey":"","enabled":"true","ringSize":"-1","consumersBeforeDispatch":"0","delayBeforeDispatch":"-1","autoDelete":"false"
            // All attributes we care about are on this form:
            // "key":"value",
            // The values we're interested in, and how they relate to QueueInfo: name, consumerCount, messageCount, messagesAcked, messagesAdded
            QueueInfo queueInfo = new QueueInfo();
            queueInfo.setHost(host);
            queueInfo.setTrimmedHost(host);
            queueInfo.setConsumers(Integer.parseInt(findValueInElement(elementJson, "consumerCount")));
            queueInfo.setDequeueCount(Integer.parseInt(findValueInElement(elementJson, "messagesAcked")));
            queueInfo.setEnqueueCount(Integer.parseInt(findValueInElement(elementJson, "messagesAdded")));
            queueInfo.setName(findValueInElement(elementJson, "name"));
            queueInfo.setSize(Integer.parseInt(findValueInElement(elementJson, "messageCount")));
            queueInfoList.add(queueInfo);
        }
        return queueInfoList;
    }

    private static List<QueueInfo> trimHostNames(List<QueueInfo> qiList) {
        // Assume that most hostname have the same suffixes
        // Start from the end of the hostname and examine last character of every hostname
        // If all hostnames have the same last character, remove that character from all hostnames
        // and save that character as suffix
        // Repeat process until a character differs on any hostname or there are no dots in the hostname
        Set<String> noDots;
        Set<String> lastChars;
        do {
            // Check last character of all hostnames
            noDots = qiList.stream().map(QueueInfo::getTrimmedHost).filter(s -> !s.contains(".")).collect(Collectors.toSet()); // All hostnames without dots
            lastChars = qiList.stream().map(QueueInfo::getTrimmedHost).map(s -> s.substring(s.length() - 1)).collect(Collectors.toSet());
            if (lastChars.size() == 1 && noDots.isEmpty()) {
                // If lastChar is not "" and
                for (QueueInfo qi : qiList) {
                    String trimmedHost = qi.getTrimmedHost();
                    if (trimmedHost.length() > 0) {
                        qi.setTrimmedHost(trimmedHost.substring(0, trimmedHost.length() - 1));
                    }
                }
            }
            // If lastChar is not "" and
        } while (lastChars.size() == 1 && noDots.isEmpty());
        return qiList;
    }

    private static List<QueueInfo> filterByQueueNames(List<QueueInfo> QIList, String queueNames) {
        if (queueNames == null || queueNames.trim().isEmpty())
            return QIList;
        List<QueueInfo> filteredQIList = new ArrayList<>();
        for (QueueInfo qi : QIList) {
            for (String matchQueueName : queueNames.split(",")) {
                if (qi.getName().equalsIgnoreCase(matchQueueName))
                    filteredQIList.add(qi);
            }
        }
        return filteredQIList;
    }

    private static List<QueueInfo> filterBySearchString(List<QueueInfo> QIList, String matchStr) {
        if (matchStr == null || matchStr.trim().isEmpty())
            return QIList;
        List<QueueInfo> filteredQIList = new ArrayList<>();
        boolean positive = true;
        if (matchStr.startsWith("!")) {
            positive = false;
            matchStr = matchStr.substring(1);
        }
        Pattern pattern = Pattern.compile(matchStr);
        for (QueueInfo qi : QIList) {
            Matcher matcher = pattern.matcher(qi.toString());
            if (matcher.find()) {
                if (positive)
                    filteredQIList.add(qi);
            } else {
                if (!positive)
                    filteredQIList.add(qi);
            }
        }
        return filteredQIList;
    }

    private static List<QueueInfo> filterByConsumerCount(List<QueueInfo> QIList, int consumerCount) {
        if (consumerCount < 0)
            return QIList;
        List<QueueInfo> filteredQIList = new ArrayList<>();
        for (QueueInfo qi : QIList) {
            if (qi.getConsumers() == consumerCount)
                filteredQIList.add(qi);
        }
        return filteredQIList;
    }

    private static List<QueueInfo> sortQueueInfo(List<QueueInfo> QIL, int sortAscending, int sortDescending) {
        if (sortAscending > 0) {
            switch (sortAscending) {
                case 1:
                    QIL.sort(Comparator.comparing(QueueInfo::getHost));
                    break;
                case 2:
                    QIL.sort(Comparator.comparing(QueueInfo::getName));
                    break;
                case 3:
                    QIL.sort(Comparator.comparingInt(QueueInfo::getSize));
                    break;
                case 4:
                    QIL.sort(Comparator.comparingInt(QueueInfo::getConsumers));
                    break;
                case 5:
                    QIL.sort(Comparator.comparingInt(QueueInfo::getEnqueueCount));
                    break;
                case 6:
                    QIL.sort(Comparator.comparingInt(QueueInfo::getDequeueCount));
                    break;
            }
        } else if (sortDescending > 0) {
            switch (sortDescending) {
                case 1:
                    QIL.sort((QueueInfo o1, QueueInfo o2) -> o2.getHost().compareTo(o1.getHost()));
                    break;
                case 2:
                    QIL.sort((QueueInfo o1, QueueInfo o2) -> o2.getName().compareTo(o1.getName()));
                    break;
                case 3:
                    QIL.sort((QueueInfo o1, QueueInfo o2) -> Integer.compare(o2.getSize(), o1.getSize()));
                    break;
                case 4:
                    QIL.sort((QueueInfo o1, QueueInfo o2) -> Integer.compare(o2.getConsumers(), o1.getConsumers()));
                    break;
                case 5:
                    QIL.sort((QueueInfo o1, QueueInfo o2) -> Integer.compare(o2.getEnqueueCount(), o1.getEnqueueCount()));
                    break;
                case 6:
                    QIL.sort((QueueInfo o1, QueueInfo o2) -> Integer.compare(o2.getDequeueCount(), o1.getDequeueCount()));
                    break;
            }
        }
        return QIL;
    }



    /* HELPER METHODS */

    private static String findValueInElement(String json, String key) {
        int keyPos = json.indexOf("\"" + key + "\":") + key.length() + 3;
        int valueStartPos = json.indexOf("\"", keyPos) + 1;
        int valueEndPos = json.indexOf("\"", valueStartPos);
        return json.substring(valueStartPos, valueEndPos);
    }

    private static String format(int size, int number) {
        return String.format("%" + size + "d, ", number);
    }

    private static String formatNoComma(int size, int number) {
        return String.format("%" + size + "d ", number);
    }

    private static String format(int size, String str) {
        return String.format("%-" + size + "s, ", "\"" + str + "\"");
    }

    private static void usage() {
        System.out.println("BrokerShell (BRS) v" + VERSION + " is made to browse and manage queues on Artemis brokers used by");
        System.out.println("ECP/EDX-installations. The broker must respond on 8161-port (console-api-port) for this tool to work.");
        System.out.println("You can run BRS locally on same host, then things will most likely work. If you want to allow for");
        System.out.println("remote access (which will enable the greatest power of BRS), then change the jolokia-access.xml to this:");
        System.out.println("    <restrict><cors><allow-origin>*</allow-origin></cors></restrict>");
        System.out.println("If you allow for remote access, then you can specify multiple hosts in the -h option, and BRS will list");
        System.out.println("queues and perform delete/purge operations on all hosts listed. With the available you can easily manage");
        System.out.println("queues on multiple brokers from a single command-line tool.");
        System.out.println();
        System.out.println("There are three main usecases for BRS:");
        System.out.println(" 1) List all queues with relevant information - the output can be used by Splunk or similar");
        System.out.println(" 2) Purge/remove queues - (be careful!)");
        System.out.println(" 3) Interactive mode for browsing and managing queues");
        System.out.println();
        System.out.println("Usage  : java -jar ekit.jar BRS [OPTIONS] QUEUE BROKER\n");
        System.out.println("\n\nThe options and arguments:");
        System.out.println(" OPTIONS       : ");
        System.out.println("                 -c<no_of_consumers>: lists queues with this number of consumers");
        System.out.println("                 -q <queuenames>    : lists queues with these names (comma-separated)");
        System.out.println("                 -s <string-search> : lists queues with info matching this regexp. Prefixing with ! will negate search.");
        System.out.println("                 -a<column>         : sort list by column (1-6) in ascending order");
        System.out.println("                 -d<column>         : sort list by column (1-6) in descending order");
        System.out.println("                 -l<no_of_lines>    : limit output to this number of lines (after filtering/sorting has been done)");
        System.out.println("                 -i                 : interactive mode - browse and manage queues");
        System.out.println("                 -r                 : remove all queues listed (use with caution)");
        System.out.println("                 -p                 : purge all queues listed (use with caution)");
        System.out.println("                 -h <host-file>     : each line specify: hostname [user password protocol] (3 last fields are optional)");
        System.out.println("                 -o <json-file>     : print queue-info to this file (json-format) (ignored in -i mode)");
        System.out.println(" BROKER        : <URL|PATH> - either specify a broker-URL, or a path for the ECP/EDX-configuration");
        System.out.println("                 URL                : Specify url on this form: http[s]://user:pass@host:port");
        System.out.println("                 PATH               : Specify either the conf-path for ECP or EDX (ex: /etc/ecp-endpoint or /etc/edx-toolbox)");
        System.out.println();
        System.out.println("\nExample 1:  List all queues on a broker");
        System.out.println("\tjava -jar ekit.jar BRS https://endpoint:password@localhost:8161");
        System.out.println("\nExample 2:  List all queues on a broker with 0 consumers");
        System.out.println("\tjava -jar ekit.jar BRS -c0 https://endpoint:password@localhost:8161");
        System.out.println("\nExample 3:  List all queues on a broker with 0 consumers and with 'test' in the name");
        System.out.println("\tjava -jar ekit.jar BRS -c0 -s test https://endpoint:password@localhost:8161");
        System.out.println("\nExample 4:  List all queues on a broker with 0 and sort by size");
        System.out.println("\tjava -jar ekit.jar BRS -c0 -a3 https://endpoint:password@localhost:8161");
        System.out.println("\nExample 5:  List all queues on a broker with 0 consumers, sort by size descending, list first 10");
        System.out.println("\tjava -jar ekit.jar BRS -c0 -d3 -l10 https://endpoint:password@localhost:8161");
        System.out.println("\nExample 6:  Interactive mode - browse and manage queues");
        System.out.println("\tjava -jar ekit.jar BRS -i https://endpoint:password@remotehost:8161");
        System.out.println("\nExample 7:  Interactive mode - browse and manage queues on multiple hosts");
        System.out.println("\tjava -jar ekit.jar BRS -i -h hosts.txt https://endpoint:password@remotehost:8161");
    }

}
