package no.statnett.ecp.eml;

import no.statnett.ecp.utils.Const;
import no.statnett.ecp.utils.DebugLog;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.StandardOpenOption;
import java.time.LocalDateTime;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.Comparator;
import java.util.List;
import java.util.Locale;
import java.util.stream.Stream;
import java.util.zip.GZIPOutputStream;

import static java.time.ZoneOffset.UTC;

public class Rotation {

    public static final String FULLTMSREGEXP = "^\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2}\\.\\d{3}Z.*";


    private File logFile;
    private long newestEcpTms = 0;
    private int noOfDaysDelayRotation;
    private DebugLog debugLog;


    public Rotation(File logFile, long newestEcpTms, int noOfDaysDelayRotation, DebugLog debugLog) {
        this.logFile = logFile;
        this.newestEcpTms = newestEcpTms;
        this.noOfDaysDelayRotation = noOfDaysDelayRotation;
        this.debugLog = debugLog;
    }

    public void deleteExcessLogs(int maxNoOfLogFiles) {
        File[] rotatedLogFiles = getRotatedLogFiles();
        int numberOfFilesToDelete = rotatedLogFiles.length - maxNoOfLogFiles;
        if (numberOfFilesToDelete > 0) {
            List<File> files = new java.util.ArrayList<>(List.of(rotatedLogFiles));
            // Sort the files by lastModified, so we delete the oldest files first
            files.sort(Comparator.comparingLong(File::lastModified));

            for (int i = 0; i < numberOfFilesToDelete; i++) {
                try {
                    Files.delete(rotatedLogFiles[i].toPath());
                    debug("Log-file " + rotatedLogFiles[i].getName() + " deleted since it's an excess log-file (limit is " + maxNoOfLogFiles + ")");
                } catch (IOException e) {
                    debug("Error occurred during deletion of log-file " + rotatedLogFiles[i].getName() + ": " + e.getMessage());
                }
            }
        }
    }

    public void compress() {
        File[] rotatedLogFiles = getRotatedLogFiles();
        // Compress logFile on the form
        // logFile-20240509.log to logFile-20240509.log.gz
        // logFile-2024W19.log to logFile-2024W19.log.gz
        // logFile-202405.log to logFile-202405.log.gz
        for (File rotatedLogFile : rotatedLogFiles) {
            if (!rotatedLogFile.getName().endsWith(".gz")) {
                gzip(rotatedLogFile);
            }
        }
    }

    public void gzip(File sourceFile) {
        File gzipFile = new File(sourceFile.getAbsolutePath() + ".gz");

        try (FileInputStream fis = new FileInputStream(sourceFile);
             FileOutputStream fos = new FileOutputStream(gzipFile);
             GZIPOutputStream gzipOS = new GZIPOutputStream(fos)) {

            byte[] buffer = new byte[1024];
            int len;
            while ((len = fis.read(buffer)) != -1) {
                gzipOS.write(buffer, 0, len);
            }

        } catch (IOException e) {
            debug("Error occurred during compression of log-file " + sourceFile.getName() + ": " + e.getMessage());
        }
        // delete source file
        try {
            Files.delete(sourceFile.toPath());
            debug("Log-file " + sourceFile.getName() + " compressed to " + gzipFile.getName() + " and then deleted");
        } catch (IOException e) {
            debug("Error occurred during deletion of log-file (excess log-file) " + sourceFile.getName() + ": " + e.getMessage());
        }


    }

    private File[] getRotatedLogFiles() {
        int dotPos = logFile.getName().indexOf(".");
        // All rotated log files have starts this way (logFile-20240509.log or logFile-2024W19.log or logFile-202405.log), this code is safe until year 3000
        String prefix = logFile.getName().substring(0, dotPos) + "-2";
        File directory = logFile.getParentFile();
        if (directory == null) {
            directory = new File(".");
        }
        if (directory.exists() && directory.isDirectory()) {
            return directory.listFiles((dir, name) -> name.startsWith(prefix));
        }
        return new File[]{};
    }


    public void rotateLogs(String rotationType) {

        // Read from beginning of logFile and find first timestamp on the form 2024-05-09T07:01:57.960Z, return this timestamp as ZonedDateTime
        ZonedDateTime firstTmsZDT = findFirstTimestampInLogFile(logFile);
        if (firstTmsZDT == null)
            return;
        // Find the last timestamp written to the log
        ZonedDateTime newestUTC = LocalDateTime.parse("" + newestEcpTms, Const.tmsNumberMillisec).atZone(UTC);
        ZonedDateTime endLoggingMinuteZDT = newestUTC.truncatedTo(ChronoUnit.MINUTES);
        // Subtract noOfDaysDelayRotation from endLoggingMinuteZDT
        endLoggingMinuteZDT = endLoggingMinuteZDT.minusDays(noOfDaysDelayRotation);
        if (endLoggingMinuteZDT.isBefore(firstTmsZDT)) {
            debug("No rotation needed for log-file " + logFile.getName() + " since endLoggingMinuteZDT - " + noOfDaysDelayRotation + " days is " + endLoggingMinuteZDT + " and is before firstTmsZDT (" + firstTmsZDT + ")");
            return;
        }

        // rotationType can be day, week or month
        String timeFormat = getTimeFormat(rotationType);
        String startTimeId = firstTmsZDT.format(DateTimeFormatter.ofPattern(timeFormat));
        String endTimeId = endLoggingMinuteZDT.format(DateTimeFormatter.ofPattern(timeFormat));
        if (!startTimeId.equals(endTimeId)) {
            // Read all lines in file and write each line to a new file with the date in the filename.
            // Discard lines that do not match timestamp-regexp in the beginng of the line
            // Each time a new currentTimeId is found, create a new file and write to this file until the next date is found
            // When endTimeId is found, write to this file until the endTimeId of the file, but keep the logname as it is
            try {
                // First rotation-file startTimeId with startTimeId, last will be the same filename as logFile with ".tmp" extension
                File rotationFile = createFile(logFile, startTimeId, endTimeId);
                String previousTimeId = null;
                Stream<String> lines = Files.lines(logFile.toPath());
                StringBuilder buffer = new StringBuilder(); // Will hold the lines to be written to the rotationFile, when buffer.length() > 1000000 (write line by line is slow)
                for (String line : (Iterable<String>) lines::iterator) {
                    if (line.length() > 24 && line.startsWith("202") && line.charAt(4) == '-') {
                        String currentTimeId = ZonedDateTime.parse(line.substring(0, 24)).format(DateTimeFormatter.ofPattern(timeFormat, new Locale("no", "NO")));
                        // If the rotation file is a tmp file, then we have reached the end of the ecp-logs (because endTimeId is reached)
                        // and we should write all the rest of the eml-file to the tmp-file (which will in the end be renamed to the original log-file)
                        // If it's not a tmp file, then we should create a new file if the currentTimeId is different from the previousTimeId
                        if (!rotationFile.getName().endsWith(".tmp") && previousTimeId != null && !currentTimeId.equals(previousTimeId)) {
                            if (buffer.length() > 0) {
                                Files.writeString(rotationFile.toPath(), buffer.toString(), StandardOpenOption.APPEND);
                                buffer.setLength(0);
                                debug("Rotation of  log-file " + rotationFile.getName() + " completed, since currentTimeId (" + currentTimeId + ") is different from previousTimeId (" + previousTimeId + ") based on timestamp " + line.substring(0, 24));
                            }
                            rotationFile = createFile(logFile, currentTimeId, endTimeId);
                        }
                        buffer.append(line).append("\n");
                        if (buffer.length() > 1000000) {
                            Files.writeString(rotationFile.toPath(), buffer.toString(), StandardOpenOption.APPEND);
                            buffer.setLength(0);
                        }
                        previousTimeId = currentTimeId;
                    }
                }
                if (buffer.length() > 0) {
                    Files.writeString(rotationFile.toPath(), buffer.toString(), StandardOpenOption.APPEND);
                    buffer.setLength(0);
                    debug("Rotation of  log-file " + rotationFile.getName() + " completed");
                }
                try {
                    // When completed - delete logFile
                    Files.delete(logFile.toPath());
                    // If lastRotationFile has a ".tmp" extension, rename it to the original logFile, else create a new file with the original logFile name
                    if (rotationFile.getName().endsWith(".tmp")) {
                        Files.move(rotationFile.toPath(), logFile.toPath());
                    } else {
                        Files.createFile(logFile.toPath());
                    }
                } catch (Throwable t) {
                    debug("Error occurred during delete or renameing of log-file " + rotationFile.getName() + " to " + logFile.getName() + " - the process will run once more next minute: " + t.getMessage());
                }
            } catch (IOException e) {
                debug("Error occurred during rotation of log-file " + logFile.getName() + ", a new attempt will begun next minute: " + e.getMessage());
            }
        }
    }

    private ZonedDateTime findFirstTimestampInLogFile(File logFile) {
        try {
            String firstLoglineWithTms = Files.lines(logFile.toPath(), StandardCharsets.UTF_8).filter(s -> s.length() > 24 && s.matches(FULLTMSREGEXP)).findFirst().orElse(null);
            if (firstLoglineWithTms != null) {
                return ZonedDateTime.parse(firstLoglineWithTms.substring(0, 24));
            } else {
                debug("No timestamp found in log-file " + logFile.getName() + " - will not rotate log-file");
            }
        } catch (IOException e) {
            debug("Error occurred during reading of log-file during findFirstTimestampInLogFile()" + logFile.getName() + " - will not rotate log-file: " + e.getMessage());
        }
        return null;
    }

    private String getTimeFormat(String rotationType) {
        String timeFormat = "yyyyMMdd";
        if (rotationType.equals("week")) {
            timeFormat = "yyyy'W'ww";
        } else if (rotationType.equals("month")) {
            timeFormat = "yyyyMM";
        }
        return timeFormat;
    }

    private File createFile(File mainLogFile, String currentTimeId, String endTimeId) throws IOException {
        int dotPos = mainLogFile.getName().indexOf("."); // Find the position of the '.' in the filename - we need this to create new filenames
        File rotationFile = new File(mainLogFile.getParent() + "/" + getNewFilename(mainLogFile, dotPos, currentTimeId, endTimeId));
        if (rotationFile.exists()) { // Delete file if it already exists, so we start with an empty file
            Files.delete(rotationFile.toPath());
        }
        rotationFile.createNewFile();
        return rotationFile;
    }

    private String getNewFilename(File logFile, int dotPos, String date, String endDate) {
        if (date.equals(endDate)) {
            return logFile.getName() + ".tmp";
        } else {
            return logFile.getName().substring(0, dotPos) + "-" + date + logFile.getName().substring(dotPos);
        }
    }

    private void debug(String message) {
       debugLog.debug(message);
    }

    public static void main(String[] args) {
        System.out.println(ZonedDateTime.parse("2024-04-28T00:00:00.821Z").format(DateTimeFormatter.ofPattern("yyyy'W'ww")));
    }

}
