package no.statnett.ecp.utils;

import jakarta.jms.*;
import org.apache.qpid.jms.JmsConnectionFactory;

import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.X509Certificate;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class EcpBroker {

    /**
     * This method will create a connection to the AMQP-broker and start it. It requires a config-map with the following keys:
     * <p>
     * url: The URL to the broker (ex: amqps://localhost:5672) - the URL is without user/pass because that's how the library wants it
     * user: The user to connect with the AMQP-broker, I think it's safe to set it even if the broker does not require it
     * password: The password to connect with the AMQP-broker, I think it's safe to set it even if the broker does not require it
     * host: The host part of the URL - yes, it founds in the URL also, but we can skip some parsing by having it here
     * port: The port part of the URL - yes, it founds in the URL also, but we can skip some parsing by having it here
     * <p>
     * The method will try to connect to the broker with TLS if that is what the url suggests (starting with amqps) otherwise
     * it will try without TLS. However, if the user has specified the URL directly, the method will try the opposite suggestion
     * also, if the first suggestion fails. For the TLS-case, the method trust ALL certificates from the server, therefore requires no truststore,
     * nor does it require a verification of the host name. This makes it easy to use. However, if a client needs to know as
     * certain which server it is connected to, then this method is not advisable to use. However, users of this method are not
     * expected to be in such a situation, the risk is entirely on the server side to allow connection.
     *
     * @param config
     * @return
     */
    public static Connection createAndStartConnection(Map<String, String> config) {
        String suggest = config.get("suggest");
        String failed = config.get("failed");
        if (suggest == null) { // User is not suggestion any URL, but referring to ecp/edx.properties - so no reason to try anything else than the url found in the properties
            if (config.get("url").startsWith("amqps")) {
                return createConnectionImpl(config, true, true);
            } else {
                return createConnectionImpl(config, false, true);
            }
        } else { // User has suggested a URL, but it can be wrong, so we must try the opposite suggestion if the first fails
            boolean tls = suggest.equals("tls"); // can be either "tls" or "no-tls" (the user have specified amqp:// or amqps://)
            if (failed == null) { // First time run, or the last time it ran it was not necessary to switch to opposite suggestion
                Connection c = createConnectionImpl(config, tls, false);
                if (c == null) {
                    config.put("failed", "true");
                    return createConnectionImpl(config, !tls, true);
                } else {
                    return c;
                }
            } else { // This can only happen if this method has been run before and returned successfully - since failed is set if the first attempt failed
                return createConnectionImpl(config, true, true);
            }
        }
    }

    private static Connection createConnectionImpl(Map<String, String> config, boolean tls, boolean exitOnFail) {
        String url = config.get("url");
        String user = config.get("user");
        String password = config.get("password");

        String modifiedUrl = url + (tls ? "?transport.verifyHost=false" : "");
        modifiedUrl = tls ? modifiedUrl.replaceAll("amqp:", "amqps:") : modifiedUrl.replaceAll("amqps:", "amqp:");

        JmsConnectionFactory factory = new JmsConnectionFactory(modifiedUrl);
        factory.setUsername(user);
        factory.setPassword(password);

        try {
            if (tls) {
                // Section which trust all certificates from the server
                SSLContext sslContext = SSLContext.getInstance("TLS");
                TrustManager[] trustManagers = new TrustManager[]{new X509TrustManager() {
                    public void checkClientTrusted(X509Certificate[] chain, String authType) {
                    }

                    public void checkServerTrusted(X509Certificate[] chain, String authType) {
                    }

                    public X509Certificate[] getAcceptedIssuers() {
                        return new X509Certificate[]{};
                    }
                }};
                sslContext.init(null, trustManagers, null);
                factory.setSslContext(sslContext);
                //  We could, if we wanted, require trustStore and specify it, to trust certain certificates only. For most
                //  users connecting remotely to an AMQP-server it is just a hassle to retrieve/copy the server's keystores.
                //        System.setProperty("transport.trustStoreLocation", config.get("keystore-file"));
                //        System.setProperty("transport.trustStorePassword", config.get("keystore-pw"));
            }
            Connection connection = factory.createConnection();
            connection.start();
            return connection;
        } catch (JMSException | NoSuchAlgorithmException | KeyManagementException e) {
            if (exitOnFail) {
                System.out.println("Connection to AMQP-server on " + config.get("host") + ":" + config.get("port") + " failed (" + e.getMessage() + ") - exiting.");
                System.exit(1);
                return null; // Only to satisfy the compiler - this line will never be reached
            } else {
                return null;
            }
        }
    }

    /**
     * This method will find the configuration for the AMQP-broker. It can take two types of input:
     * 1) A URL to the broker (ex: amqps://localhost:5672)
     * 2) A path to a directory containing ecp.properties or edx.properties
     * Based on the input, the method will return a map with the following data:
     * <p>
     * url: The URL to the broker (ex: amqps://localhost:5672) - the URL is without user/pass because that's how the library wants it
     * user: The user to connect with the AMQP-broker, I think it's safe to set it even if the broker does not require it
     * password: The password to connect with the AMQP-broker, I think it's safe to set it even if the broker does not require it
     * host: The host part of the URL - yes, it founds in the URL also, but we can skip some parsing by having it here
     * port: The port part of the URL - yes, it founds in the URL also, but we can skip some parsing by having it here
     * Only for scenario 1) - URL to the broker:
     * suggest: If the user has specified the URL directly, this key will be set to "tls" or "no-tls" based on the URL. This
     * is used to suggest the opposite URL if the first attempt fails (so it doesn't matter as much if the wrong scheme
     * is given in the URL)
     * Only for scenario 2) - Path to a directory containing ecp.properties or edx.properties:
     * ecp: If the path was to a directory containing ecp.properties, this key will be set to "true"
     * edx: If the path was to a directory containing edx.properties, this key will be set to "true"
     * yml: If the path was to a directory containing edx.properties, this key will be set to the path of the edx.yml file
     * logs: The path to the log file for the endpoint
     *
     * @param urlPathArg
     * @return
     * @throws IOException
     */
    public static Map<String, String> findBrokerConfig(String urlPathArg, boolean amqpRequired) throws IOException {
        Map<String, String> configMap = new HashMap<>();
        if ((urlPathArg.startsWith("amqp") && amqpRequired) || (urlPathArg.startsWith("http") && !amqpRequired)) { // The URL has been given directly - no need to parse ecp/edx.properties
            // The url may look like this: amqp://endpoint:password@h1-a-nmms-apt01.statnett.no:5672 (or starting with amqps)
            // OR
            // The url may look like this: http://endpoint:password@h1-a-nmms-apt01.statnett.no:8161 (or starting with https)
            // We must remove user and password from the URL and put them in the configMap
            int userStart = urlPathArg.indexOf("//") + 2;
            int passStart = urlPathArg.indexOf(":", userStart) + 1;
            int hostStart = urlPathArg.indexOf("@") + 1;
            int portStart = urlPathArg.indexOf(":", Math.max(userStart, hostStart)) + 1; // should always find a port number
            try {
                if (portStart > 0) {
                    Integer.parseInt(urlPathArg.substring(portStart));
                }
            } catch (NumberFormatException e) {
                System.out.println("This part '" + urlPathArg.substring(portStart) + "' was assumed to be the port number, but is not a valid port number");
                if (amqpRequired)
                    System.out.println("The BROKER URL must be like this 'amqp[s]://[user:pass@]host:port' (ex: amqp://localhost:5672) - exiting.");
                else
                    System.out.println("The BROKER URL must be like this 'http[s]://[user:pass@]host:port' (ex: https://endpoint:password@localhost:8161) - exiting.");
                System.exit(1);

            }
            if (userStart > 6 && passStart > userStart && hostStart > passStart && portStart > hostStart) {
                configMap.put("user", urlPathArg.substring(userStart, passStart - 1));
                configMap.put("password", urlPathArg.substring(passStart, hostStart - 1));
                configMap.put("host", urlPathArg.substring(hostStart, portStart - 1));
                configMap.put("port", urlPathArg.substring(portStart));
                configMap.put("url", urlPathArg.substring(0, userStart) + urlPathArg.substring(hostStart)); // Remove user and pass from URL
                if (urlPathArg.startsWith("amqps") || urlPathArg.startsWith("https")) {
                    configMap.put("suggest", "tls");
                } else {
                    configMap.put("suggest", "no-tls");
                }
            } else if (userStart > 6 && portStart > userStart) {
                configMap.put("host", urlPathArg.substring(userStart, portStart - 1));
                configMap.put("port", urlPathArg.substring(portStart));
                configMap.put("url", urlPathArg); // Remove user and pass from URL
            } else {
                if (amqpRequired)
                    System.out.println("BROKER URL must be like this 'amqp[s]://[user:pass@]host:port' (ex: amqp://localhost:5672) - exiting.");
                else
                    System.out.println("BROKER URL must be like this 'http[s]://[user:pass@]host:port' (ex: https://endpoint:password@localhost:8161) - exiting.");
                System.exit(1);
            }
        } else {
            String url = "";
            File f = new File(urlPathArg);
            if (f.isDirectory()) {
                File ecpPropFile = new File(urlPathArg + "/" + "ecp.properties");
                if (ecpPropFile.exists()) { // ECP-config
                    Options.checkFile(ecpPropFile.getAbsolutePath(), false);
                    String host = findHost(urlPathArg);
                    configMap.put("host", host);
                    List<String> lines = Files.readAllLines(ecpPropFile.toPath());
                    Map<String, String> map = makeMap(lines);
                    String auth = getProp(map, "internalBroker.useAuthentication", "true", true);
                    if (auth.equals("true")) {
                        url += "amqps://" + host + ":";
                    } else {
                        url += "amqp://" + host + ":";
                    }
                    String port = getProp(map, "internalBroker.amqp.port", "5672");
                    configMap.put("port", port);
                    url += port;
                    configMap.put("url", url);
                    configMap.put("user", getProp(map, "internalBroker.auth.user", "endpoint"));
                    configMap.put("password", getProp(map, "internalBroker.auth.password", "password"));
                    configMap.put("ecp", "true");
                    String ecpLog = getProp(map, "logging.file.name", "/var/log/ecp-endpoint/ecp.log", false);
                    Options.checkFile(ecpLog, false);
                    configMap.put("log", ecpLog);
                } else {
                    File edxPropFile = new File(urlPathArg + "/" + "edx.properties");
                    if (edxPropFile.exists()) { // EDX-config
                        Options.checkFile(edxPropFile.getAbsolutePath(), false);
                        String host = findHost(urlPathArg);
                        configMap.put("host", host);
                        List<String> lines = Files.readAllLines(edxPropFile.toPath());
                        Map<String, String> map = makeMap(lines);
                        String auth = getProp(map, "internalBroker.useAuthentication", "true", true);
                        if (auth.equals("true")) {
                            url += "amqps://" + host + ":";
                        } else {
                            url += "amqp://" + host + ":";
                        }
                        String port = getProp(map, "internalBroker.amqp.port", "5672");
                        configMap.put("port", port);
                        url += port;
                        configMap.put("url", url);
                        configMap.put("user", getProp(map, "internalBroker.auth.user", "toolbox"));
                        configMap.put("password", getProp(map, "internalBroker.auth.password", "password"));
                        configMap.put("edx", "true");
                        configMap.put("yml", urlPathArg + "/" + "edx.yml");
                        String edxLogs = getProp(map, "logging.file.name", "/var/log/edx-toolbox/edx.log", false);
                        Options.checkFile(edxLogs, false);
                        configMap.put("log", edxLogs);

                    } else {
                        throw new RuntimeException("Did not find ecp.properties nor edx.properties in " + urlPathArg);
                    }
                }
            } else {
                throw new RuntimeException("Argument " + urlPathArg + " is not a directory, if you intend it to be a URL make sure it starts with 'amqp'");
            }
        }
        return configMap;
    }

    private static String findHost(String urlPathArg) {
        try {
            return Files.readString(Path.of(urlPathArg + "/host")).trim();
        } catch (Throwable t) {
            return "localhost";
        }
    }

    private static String getProp(Map<String, String> map, String name, String defaultValue) {
        return getProp(map, name, defaultValue, false);
    }

    private static String getProp(Map<String, String> map, String name, String defaultValue, boolean toLowerCase) {
        if (map.get(name) != null) {
            return toLowerCase ? map.get(name).trim().toLowerCase() : map.get(name).trim();
        } else {
            return defaultValue;
        }
    }

    private static Map<String, String> makeMap(List<String> lines) {
        Map map = new HashMap();
        for (String line : lines) {
            if (!line.trim().startsWith("#") && line.contains("=") && line.split("=").length == 2) {
                map.put(line.split("=")[0].trim(), line.split("=")[1].trim());
            }
        }
        return map;
    }

    public static String buildSelector(String messageType, String TTL, String senderCode, String receiverCode, String internalType) {
        String selector = null;
        if (messageType != null) {
            selector = "(messageType = '" + messageType + "' OR businessType = '" + messageType + "' OR EDX_MESSAGE_TYPE = '" + messageType + "')";
        }
        if (TTL != null && Long.parseLong(TTL) > 0) {
            long tms = System.currentTimeMillis() - (Long.parseLong(TTL) * 1000L);
            String newSelector = "(JMSTimestamp < " + tms + " OR generated < " + tms + ")";
//            String newSelector = "convert_string_expressions:(JMSTimestamp < " + tms + " OR generated < " + tms + ")";
//            String newSelector = "convert_string_expressions:generated < " + tms + "";
            selector = appendSelector(selector, newSelector);
        }
        if (senderCode != null && !senderCode.equals("ALL")) {
            String newSelector = "(senderCode = '" + senderCode + "' OR sender = '" + senderCode + "' OR EDX_SENDER = '" + senderCode + "')";
            selector = appendSelector(selector, newSelector);
        }
        if (receiverCode != null) {
            String newSelector = "(receiverCode = '" + receiverCode + "' OR receiver = '" + receiverCode + "' OR EDX_RECEIVER_CODE = '" + receiverCode + "')";
            selector = appendSelector(selector, newSelector);
        }
        if (internalType != null) {
            String newSelector = "(internalType = '" + internalType + "' OR EDX_MESSAGE_KIND = '" + internalType + "')";
            selector = appendSelector(selector, newSelector);
        }

        return selector;
    }

    private static String appendSelector(String selector, String TTLSelect) {
        selector = (selector == null ? TTLSelect : selector + " AND " + TTLSelect);
        return selector;
    }

}
